@@ -1,12 +1,46 @@
+"""
+MIT License
+
+Copyright (c) 2017-present Devoxin
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in all
+copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+SOFTWARE.
+"""
+import logging
+
 from .events import NodeConnectedEvent, NodeDisconnectedEvent
 from .node import Node
 
+_log = logging.getLogger(__name__)
+DEFAULT_REGIONS = {
+    'asia': ('hongkong', 'singapore', 'sydney', 'japan', 'southafrica', 'india'),
+    'eu': ('eu', 'amsterdam', 'frankfurt', 'russia', 'london'),
+    'us': ('us', 'brazil')
+}
+
 
 class NodeManager:
     """Represents the node manager that contains all lavalink nodes.
 
+    len(x):
+        Returns the total number of nodes.
     iter(x):
-        Returns an iterator of all the nodes cached.
+        Returns an iterator of all the stored nodes.
 
     Attributes
     ----------
@@ -18,14 +52,11 @@ class NodeManager:
     def __init__(self, lavalink, regions: dict):
         self._lavalink = lavalink
         self._player_queue = []
-
         self.nodes = []
+        self.regions = regions or DEFAULT_REGIONS
 
-        self.regions = regions or {
-            'asia': ('hongkong', 'singapore', 'sydney', 'japan', 'southafrica', 'india'),
-            'eu': ('eu', 'amsterdam', 'frankfurt', 'russia', 'london'),
-            'us': ('us', 'brazil')
-        }
+    def __len__(self):
+        return len(self.nodes)
 
     def __iter__(self):
         for n in self.nodes:
@@ -38,7 +69,7 @@ class NodeManager:
 
     def add_node(self, host: str, port: int, password: str, region: str,
                  resume_key: str = None, resume_timeout: int = 60, name: str = None,
-                 reconnect_attempts: int = 3):
+                 reconnect_attempts: int = 3, filters: bool = False, ssl: bool = False):
         """
         Adds a node to Lavalink's node manager.
 
@@ -54,32 +85,43 @@ class NodeManager:
             The region to assign this node to.
         resume_key: Optional[:class:`str`]
             A resume key used for resuming a session upon re-establishing a WebSocket connection to Lavalink.
-            Defaults to `None`.
+            Defaults to ``None``.
         resume_timeout: Optional[:class:`int`]
             How long the node should wait for a connection while disconnected before clearing all players.
-            Defaults to `60`.
+            Defaults to ``60``.
         name: Optional[:class:`str`]
-            An identifier for the node that will show in logs. Defaults to `None`.
+            An identifier for the node that will show in logs. Defaults to ``None``.
         reconnect_attempts: Optional[:class:`int`]
             The amount of times connection with the node will be reattempted before giving up.
-            Set to `-1` for infinite. Defaults to `3`.
+            Set to `-1` for infinite. Defaults to ``3``.
+        filters: Optional[:class:`bool`]
+            Whether to use the new ``filters`` op. This setting currently only applies to development
+            Lavalink builds, where the ``equalizer`` op was swapped out for the broader ``filters`` op which
+            offers more than just equalizer functionality. Ideally, you should only change this setting if you
+            know what you're doing, as this can prevent the effects from working.
+        ssl: Optional[:class:`bool`]
+            Whether to use SSL for the node. SSL will use ``wss`` and ``https``, instead of ``ws`` and ``http``,
+            respectively. Your node should support SSL if you intend to enable this, either via reverse proxy or
+            other methods. Only enable this if you know what you're doing.
         """
-        node = Node(self, host, port, password, region, resume_key, resume_timeout, name, reconnect_attempts)
+        node = Node(self, host, port, password, region, resume_key, resume_timeout, name, reconnect_attempts, filters, ssl)
         self.nodes.append(node)
 
-        self._lavalink._logger.info('[NODE-{}] Successfully added to Node Manager'.format(node.name))
+        _log.info('Added node \'%s\'', node.name)
 
     def remove_node(self, node: Node):
         """
         Removes a node.
 
+        Make sure you have called :func:`Node.destroy` to close any open WebSocket connection.
+
         Parameters
         ----------
         node: :class:`Node`
             The node to remove from the list.
         """
         self.nodes.remove(node)
-        self._lavalink._logger.info('[NODE-{}] Successfully removed Node'.format(node.name))
+        _log.info('Removed node \'%s\'', node.name)
 
     def get_region(self, endpoint: str):
         """
@@ -117,7 +159,7 @@ class NodeManager:
         Parameters
         ----------
         region: Optional[:class:`str`]
-            The region to find a node in. Defaults to `None`.
+            The region to find a node in. Defaults to ``None``.
 
         Returns
         -------
@@ -145,11 +187,10 @@ class NodeManager:
         node: :class:`Node`
             The node that has just connected.
         """
-        self._lavalink._logger.info('[NODE-{}] Successfully established connection'.format(node.name))
-
         for player in self._player_queue:
             await player.change_node(node)
-            self._lavalink._logger.debug('[NODE-{}] Successfully moved {}'.format(node.name, player.guild_id))
+            original_node_name = player._original_node.name if player._original_node else '[no node]'
+            _log.debug('Moved player %d from node \'%s\' to node \'%s\'', player.guild_id, original_node_name, node.name)
 
         if self._lavalink._connect_back:
             for player in node._original_players:
@@ -172,16 +213,19 @@ class NodeManager:
         reason: :class:`str`
             The reason why the node was disconnected.
         """
-        self._lavalink._logger.warning('[NODE-{}] Disconnected with code {} and reason {}'.format(node.name, code,
-                                                                                                  reason))
+        for player in node.players:
+            try:
+                await player.node_unavailable()
+            except:  # noqa: E722 pylint: disable=bare-except
+                _log.exception('An error occurred whilst calling player.node_unavailable()')
+
         await self._lavalink._dispatch_event(NodeDisconnectedEvent(node, code, reason))
 
         best_node = self.find_ideal_node(node.region)
 
         if not best_node:
             self._player_queue.extend(node.players)
-            self._lavalink._logger.error('Unable to move players, no available nodes! '
-                                         'Waiting for a node to become available.')
+            _log.error('Unable to move players, no available nodes! Waiting for a node to become available.')
             return
 
         for player in node.players:
